#ifndef AMDIS_BOXSMOOTHER_H
#define AMDIS_BOXSMOOTHER_H

#include "SmootherBase.h"
#include "MemoryManager.h"
#include "DOFMatrix.h"
#include "SparseVector.h"
#include "MatrixVector.h"
#include "Global.h"
#include "GaussElimination.h"
#include "FiniteElemSpace.h"
#include "GaussElimination.h"
#include <set>

namespace AMDiS {

  class BoxSmoother 
    : public SmootherBase<Matrix<DOFMatrix*>, 
			  Vector<SparseVector<double>*>,
			  Vector< ::std::set<DegreeOfFreedom>*> >
  {
  public:
    MEMORY_MANAGED(BoxSmoother);

    class Creator 
      : public SmootherCreator<Matrix<DOFMatrix*>, 
			       Vector<SparseVector<double>*>,
			       Vector< ::std::set<DegreeOfFreedom>*> >
    {
    public:
      MEMORY_MANAGED(Creator);

      SmootherBase<Matrix<DOFMatrix*>, 
		   Vector<SparseVector<double>*>,
		   Vector< ::std::set<DegreeOfFreedom>*> >*
      create() {
	return NEW BoxSmoother(this->name);
      };
    };

    BoxSmoother(::std::string name)
      : SmootherBase<Matrix<DOFMatrix*>, 
		     Vector<SparseVector<double>*>,
		     Vector< ::std::set<DegreeOfFreedom>*> >(name),
	boxMatrix_(NULL),
	boxSolution_(NULL),
	boxRHS_(NULL),
	numComponents_(0),
	sumDOFs_(0)
    {
      directSolver_ = 
	NEW GaussElimination<Matrix<double>, Vector<double> >();
    };
 
    virtual ~BoxSmoother()
    {
      if(boxMatrix_) DELETE boxMatrix_;
      if(boxSolution_) DELETE boxSolution_;
      if(boxRHS_) DELETE boxRHS_;
      DELETE directSolver_;
    };

    void smooth(Matrix<DOFMatrix*>                       *matrix,
		Vector<SparseVector<double>*>            *solution,
		Vector<SparseVector<double>*>            *rhs,
		int                                       iterations,
		const Vector< ::std::set<DegreeOfFreedom>*> &dofSet)
    {
      FUNCNAME("BoxSmoother::smooth()");

      int i, j, k;

      // allocate memory
      if(!boxMatrix_) {
	numComponents_ = solution->getSize();
	numDOFs_.resize(numComponents_);
	sumDOFs_ = 0;
	for(i = 0; i < numComponents_; i++) {
	  numDOFs_[i] = (*solution)[i]->getFESpace()->getBasisFcts()->getNumber();
	  sumDOFs_ += numDOFs_[i];
	}

	boxMatrix_ = NEW Matrix<double>(sumDOFs_, sumDOFs_);
	boxSolution_ = NEW Vector<double>(sumDOFs_);
	boxRHS_ = NEW Vector<double>(sumDOFs_);
      }

      // create local equation system
      int boxSize = 0;
      ::std::vector<int> offset(numComponents_, 0);
      ::std::vector< ::std::map<DegreeOfFreedom, int> > indexOfDOF(numComponents_);
      ::std::set<DegreeOfFreedom>::const_iterator dofIt, dofBegin, dofEnd;
      ::std::vector<MatEntry>::iterator rowIt, rowBegin, rowEnd;

      int pos;
      for(i = 0; i < numComponents_; i++) {
	if(dofSet[i]) {
	  offset[i] = boxSize;
	  boxSize += numDOFs_[i];
	  dofBegin = dofSet[i]->begin();
	  dofEnd = dofSet[i]->end();
	  pos = 0;
	  for(dofIt = dofBegin; dofIt != dofEnd; ++dofIt) {
	    indexOfDOF[i][*dofIt] = pos + 1;
// 	    MSG("%d : %d, %d \n", *dofIt, i, pos);
	    ++pos;
	  }
	}
      }

      for(i = 0; i < boxSize; i++) {
	for(j = 0; j < boxSize; j++) {
	  (*boxMatrix_)[i][j] = 0.0;
	}
	(*boxRHS_)[i] = 0.0;
      }

      pos = 0;
      for(i = 0; i < numComponents_; i++) {
	if(dofSet[i]) {
	  dofBegin = dofSet[i]->begin();
	  dofEnd = dofSet[i]->end();
	  for(dofIt = dofBegin; dofIt != dofEnd; ++dofIt) {

	    // fill box rhs
	    (*boxRHS_)[pos] = (*((*rhs)[i]))[*dofIt];

	    // fill box matrix
	    for(j = 0; j < numComponents_; j++) {
	      if((*matrix)[i][j] != NULL) {
		rowBegin = (*((*matrix)[i][j]))[*dofIt].begin();
		rowEnd = (*((*matrix)[i][j]))[*dofIt].end();
		for(rowIt = rowBegin; rowIt != rowEnd; ++rowIt) {
		  if(rowIt->col == DOFMatrix::UNUSED_ENTRY) continue;
		  if(rowIt->col == DOFMatrix::NO_MORE_ENTRIES) break;
		  int ind = indexOfDOF[j][rowIt->col] - 1;
		  if(ind != -1) {
		    (*boxMatrix_)[pos][offset[j] + ind] = rowIt->entry;
		  } else {
		    (*boxRHS_)[pos] -= 
		      rowIt->entry * (*((*solution)[j]))[rowIt->col];
		  }
		}
	      } else {
		for(k = 0; k < numDOFs_[j]; k++) {
		  (*boxMatrix_)[pos][offset[j] + k] = 0.0;
		}
	      }
	    }

	    ++pos;
	  }
	}
      }

//       int a,b;

//       for(i = 0; i < numComponents_; i++) {
// 	MSG("DOFs %d: ", i);
// 	dofBegin = dofSet[i]->begin();
// 	dofEnd = dofSet[i]->end();
// 	for(dofIt = dofBegin; dofIt != dofEnd; ++dofIt) {
// 	  MSG("%d ", *dofIt);
// 	}
// 	MSG("\n");
//       }

//       for(a = 0; a < numComponents_; a++) {
// 	for(b = 0; b < numComponents_; b++) {
// 	  MSG("matrix[%d][%d]:\n", a, b);
// 	  if((*matrix)[a][b]) {
// 	    (*matrix)[a][b]->print();
// 	    MSG("solution[%d]\n", b);
// 	    (*solution)[b]->print();
// 	    MSG("rhs[%d]\n", a);
// 	    (*rhs)[a]->print();
// 	  }
// 	}
//       }

//       MSG("matrix:\n");
//       for(a = 0; a < boxSize; a++) {
// 	for(b = 0; b < boxSize; b++) {
// 	  MSG("%e ", (*boxMatrix_)[a][b]);
// 	}
// 	MSG("\n");
//       }

//       MSG("rhs:\n");
//       for(a = 0; a < boxSize; a++) {
// 	MSG("%e\n", (*boxRHS_)[a]);
//       }

      // apply direct solver
      directSolver_->solve(*boxMatrix_, *boxRHS_, *boxSolution_, boxSize);

//       MSG("solution:\n");
//       for(a = 0; a < boxSize; a++) {
// 	MSG("%e\n", (*boxSolution_)[a]);
//       }

//       WAIT_REALLY;

      // copy solution
      pos = 0;
      for(i = 0; i < numComponents_; i++) {
	dofBegin = dofSet[i]->begin();
	dofEnd = dofSet[i]->end();
	for(dofIt = dofBegin; dofIt != dofEnd; ++dofIt) {
	  (*((*solution)[i]))[*dofIt] = (*boxSolution_)[pos];	    
	  ++pos;
	}
      }
    };

  protected:
    Matrix<double> *boxMatrix_;
    Vector<double> *boxSolution_;
    Vector<double> *boxRHS_;
    int numComponents_;
    ::std::vector<int> numDOFs_;
    int sumDOFs_;
    DirectSolverInterface<Matrix<double>, Vector<double> > *directSolver_;
  };
}

#endif
